from lark import Lark, Transformer
from mel_ast import *


parser = Lark('''
    %import common.NUMBER
    %import common.CNAME
    %import common.NEWLINE
    %import common.WS
    %import common.CPP_COMMENT
    %import common.C_COMMENT
    %import common.ESCAPED_STRING

    %ignore WS

    COMMENT: CPP_COMMENT | C_COMMENT
    %ignore COMMENT
    
    BOOL: "true" | "false"
    literal: NUMBER | ESCAPED_STRING | BOOL
    ident: CNAME

    ?group: literal
        | ident
        | func_call
        | "(" expr ")"

    ?mult: group
        | mult "*" group    -> mul
        | mult "/" group    -> div
        | mult "%" group    -> mod

    ?add: mult
        | add "+" mult      -> add
        | add "-" mult      -> sub
        
    ?compare1: add
        | add ">" add       -> gt
        | add ">=" add      -> ge
        | add "<" add       -> lt
        | add "<=" add      -> le
        
    ?compare2: compare1
        | compare1 "==" compare1      -> eq
        | compare1 "!=" compare1      -> ne
        
    ?and: compare2
        | and "&&" compare2
        
    ?or: and
        | or "||" and
        
    ?expr: or
        
    func_call: ident "(" (expr ("," expr)*)? ")"

    none_stmt:   -> stmt_list
    none_expr:   -> true
    
    ?type: ident
    ?var_decl: ident
        | ident "=" expr     -> assign
    vars_decl: type var_decl ("," var_decl)*
    
    param_decl: type ident  -> vars_decl
    func_decl: type ident "(" (param_decl ("," param_decl)*)? ")" "{" stmt_list "}"
    
    ?stmt1: ident "=" expr   -> assign
        | "return" expr      -> return
        | func_call
        | vars_decl
    ?stmt2: "{" stmt_list "}"
        | "while" "(" expr ")" stmt              -> while
        | "if" "(" expr ")" stmt ("else" stmt)?  -> if
        | "for" "(" (stmt1 | none_stmt) ";" (expr | none_expr) ";" (stmt1 | none_stmt) ")" (stmt | ";" none_stmt)  -> for
        | func_decl
    ?stmt: stmt1 ";" | stmt2
    
    ?stmt_list: (stmt ";"*)*

    ?prog: stmt_list

    ?start: prog
''', start="start")


'''
class MelASTBuilder0(Transformer):
    def literal(self, arg: str) -> LiteralNode:
        return LiteralNode(arg)

    def assign(self, ident: IdentNode, expr: ExprNode) -> AssignNode:
        return AssignNode(ident, expr)

    def while(self, cond: IdentNode, body: StmtNode) -> WhileNode:
        return WhileNode(cond, body)

    def add(self, arg1: ExprNode, arg2: ExprNode) -> ExprNode:
        return BinOpNode(BinOp.ADD, arg1, arg2)

    def sub(self, arg1: ExprNode, arg2: ExprNode) -> ExprNode:
        return BinOpNode(BinOp.SUB, arg1, arg2)
'''

class MelASTBuilder(Transformer):
    def _call_userfunc(self, tree, new_children=None):
        # Assumes tree is already transformed
        children = new_children if new_children is not None else tree.children
        try:
            f = getattr(self, tree.data)
        except AttributeError:
            return self.__default__(tree.data, children, tree.meta)
        else:
            return f(*children)

    def __getattr__(self, item):
        if isinstance(item, str) and item.upper() == item:
            return lambda x: x

        if item == 'true':
            return lambda *args: LiteralNode('true')

        if item in ('mul', 'div', 'add', 'sub',
                    'gt', 'ge', 'lt', 'le', 'eq', 'ne',
                    'and', 'or'):
            def get_bin_op_node(arg1, arg2):
                op = BinOp[item.upper()]
                return BinOpNode(op, arg1, arg2)
            return get_bin_op_node
        else:
            def get_node(*args):
                cls = eval(''.join(x.capitalize() or '_' for x in item.split('_')) + 'Node')
                return cls(*args)
            return get_node


def parse(prog: str)->StmtListNode:
    prog = parser.parse(str(prog))
    prog = MelASTBuilder().transform(prog)
    return prog
